@@ -4,6 +4,7 @@
 // license that can be found in the LICENSE file.
 #include "tools/benchmark/benchmark_codec_webp.h"
 
+#include <jxl/cms.h>
 #include <stdint.h>
 #include <string.h>
 #include <webp/decode.h>
@@ -15,19 +16,23 @@
 #include "lib/extras/time.h"
 #include "lib/jxl/base/compiler_specific.h"
 #include "lib/jxl/base/data_parallel.h"
-#include "lib/jxl/base/padded_bytes.h"
 #include "lib/jxl/base/span.h"
-#include "lib/jxl/base/thread_pool_internal.h"
+#include "lib/jxl/base/status.h"
 #include "lib/jxl/codec_in_out.h"
 #include "lib/jxl/dec_external_image.h"
-#include "lib/jxl/enc_color_management.h"
 #include "lib/jxl/enc_external_image.h"
 #include "lib/jxl/enc_image_bundle.h"
 #include "lib/jxl/image.h"
 #include "lib/jxl/image_bundle.h"
 #include "lib/jxl/sanitizers.h"
+#include "tools/thread_pool_internal.h"
 
-namespace jxl {
+namespace jpegxl {
+namespace tools {
+
+using ::jxl::ImageBundle;
+using ::jxl::ImageMetadata;
+using ::jxl::ThreadPool;
 
 // Sets image data from 8-bit sRGB pixel array in bytes.
 // Amount of input bytes per pixel must be:
@@ -37,7 +42,7 @@ Status FromSRGB(const size_t xsize, const size_t ysize, const bool is_gray,
                 const JxlEndianness endianness, const uint8_t* pixels,
                 const uint8_t* end, ThreadPool* pool, ImageBundle* ib) {
   const ColorEncoding& c = ColorEncoding::SRGB(is_gray);
-  const size_t bits_per_sample = (is_16bit ? 2 : 1) * kBitsPerByte;
+  const size_t bits_per_sample = (is_16bit ? 2 : 1) * jxl::kBitsPerByte;
   const uint32_t num_channels = (is_gray ? 1 : 3) + (has_alpha ? 1 : 0);
   JxlDataType data_type = is_16bit ? JXL_TYPE_UINT16 : JXL_TYPE_UINT8;
   JxlPixelFormat format = {num_channels, data_type, endianness, 0};
@@ -85,9 +90,9 @@ class WebPCodec : public ImageCodec {
   }
 
   Status Compress(const std::string& filename, const CodecInOut* io,
-                  ThreadPoolInternal* pool, std::vector<uint8_t>* compressed,
+                  ThreadPool* pool, std::vector<uint8_t>* compressed,
                   jpegxl::tools::SpeedStats* speed_stats) override {
-    const double start = Now();
+    const double start = jxl::Now();
     const ImageBundle& ib = io->Main();
 
     if (ib.HasAlpha() && ib.metadata()->GetAlphaBits() > 8) {
@@ -99,8 +104,8 @@ class WebPCodec : public ImageCodec {
     ImageBundle store(&metadata);
     const ImageBundle* transformed;
     const ColorEncoding& c_desired = ColorEncoding::SRGB(false);
-    JXL_RETURN_IF_ERROR(TransformIfNeeded(ib, c_desired, GetJxlCms(), pool,
-                                          &store, &transformed));
+    JXL_RETURN_IF_ERROR(jxl::TransformIfNeeded(
+        ib, c_desired, *JxlGetDefaultCms(), pool, &store, &transformed));
     size_t xsize = ib.oriented_xsize();
     size_t ysize = ib.oriented_ysize();
     size_t stride = xsize * num_chans;
@@ -153,14 +158,14 @@ class WebPCodec : public ImageCodec {
     } else {
       return false;
     }
-    const double end = Now();
+    const double end = jxl::Now();
     speed_stats->NotifyElapsed(end - start);
     return true;
   }
 
   Status Decompress(const std::string& filename,
-                    const Span<const uint8_t> compressed,
-                    ThreadPoolInternal* pool, CodecInOut* io,
+                    const Span<const uint8_t> compressed, ThreadPool* pool,
+                    CodecInOut* io,
                     jpegxl::tools::SpeedStats* speed_stats) override {
     WebPDecoderConfig config;
 #ifdef MEMORY_SANITIZER
@@ -177,11 +182,11 @@ class WebPCodec : public ImageCodec {
     buf->colorspace = MODE_RGBA;
     const uint8_t* webp_data = compressed.data();
     const int webp_size = compressed.size();
-    const double start = Now();
+    const double start = jxl::Now();
     if (WebPDecode(webp_data, webp_size, &config) != VP8_STATUS_OK) {
       return JXL_FAILURE("WebPDecode failed");
     }
-    const double end = Now();
+    const double end = jxl::Now();
     speed_stats->NotifyElapsed(end - start);
     JXL_CHECK(buf->u.RGBA.stride == buf->width * 4);
 
@@ -191,7 +196,7 @@ class WebPCodec : public ImageCodec {
     const uint8_t* data_end = data_begin + buf->width * buf->height * 4;
     // The image data is initialized by libwebp, which we are not instrumenting
     // with msan.
-    msan::UnpoisonMemory(data_begin, data_end - data_begin);
+    jxl::msan::UnpoisonMemory(data_begin, data_end - data_begin);
     if (io->metadata.m.color_encoding.IsGray() != is_gray) {
       // TODO(lode): either ensure is_gray matches what the color profile says,
       // or set a correct color profile, e.g.
@@ -206,7 +211,6 @@ class WebPCodec : public ImageCodec {
                                data_begin, data_end, pool, &io->Main());
     WebPFreeDecBuffer(buf);
     JXL_RETURN_IF_ERROR(ok);
-    io->dec_pixels = buf->width * buf->height;
     return true;
   }
 
@@ -227,7 +231,9 @@ class WebPCodec : public ImageCodec {
                           std::vector<uint8_t>* compressed) {
     compressed->clear();
     WebPConfig config;
-    WebPConfigInit(&config);
+    if (!WebPConfigInit(&config)) {
+      return JXL_FAILURE("WebPConfigInit failed");
+    }
     JXL_ASSERT(!lossless_ || !near_lossless_);  // can't have both
     config.lossless = lossless_;
     config.quality = quality;
@@ -242,7 +248,9 @@ class WebPCodec : public ImageCodec {
     JXL_CHECK(WebPValidateConfig(&config));
 
     WebPPicture pic;
-    WebPPictureInit(&pic);
+    if (!WebPPictureInit(&pic)) {
+      return JXL_FAILURE("WebPPictureInit failed");
+    }
     pic.width = static_cast<int>(xsize);
     pic.height = static_cast<int>(ysize);
     pic.writer = &WebPStringWrite;
@@ -250,9 +258,13 @@ class WebPCodec : public ImageCodec {
     pic.custom_ptr = compressed;
 
     if (num_chans == 3) {
-      WebPPictureImportRGB(&pic, srgb.data(), 3 * xsize);
+      if (!WebPPictureImportRGB(&pic, srgb.data(), 3 * xsize)) {
+        return JXL_FAILURE("WebPPictureImportRGB failed");
+      }
     } else {
-      WebPPictureImportRGBA(&pic, srgb.data(), 4 * xsize);
+      if (!WebPPictureImportRGBA(&pic, srgb.data(), 4 * xsize)) {
+        return JXL_FAILURE("WebPPictureImportRGBA failed");
+      }
     }
 
     // WebP encoding may fail, for example, if the image is more than 16384
@@ -261,7 +273,7 @@ class WebPCodec : public ImageCodec {
     WebPPictureFree(&pic);
     // Compressed image data is initialized by libwebp, which we are not
     // instrumenting with msan.
-    msan::UnpoisonMemory(compressed->data(), compressed->size());
+    jxl::msan::UnpoisonMemory(compressed->data(), compressed->size());
     return ok;
   }
 
@@ -276,4 +288,5 @@ ImageCodec* CreateNewWebPCodec(const BenchmarkArgs& args) {
   return new WebPCodec(args);
 }
 
-}  // namespace jxl
+}  // namespace tools
+}  // namespace jpegxl
